@@ -1,5 +1,5 @@
 #
-# $Id: SinFP.pm,v 1.8.2.23 2005/06/20 21:43:45 gomor Exp $
+# $Id: SinFP.pm,v 1.8.2.29 2006/03/13 12:28:36 gomor Exp $
 #
 
 package Net::SinFP;
@@ -7,7 +7,7 @@ package Net::SinFP;
 use strict;
 use warnings;
 
-our $VERSION = '0.92';
+our $VERSION = '1.00';
 
 require Exporter;
 require Class::Gomor::Hash;
@@ -20,8 +20,6 @@ our @EXPORT_OK = qw(
    SIGNATURE_TYPE_EXACT
    SIGNATURE_TYPE_HEURISTIC1
    SIGNATURE_TYPE_HEURISTIC2
-
-   SIGNATURE_TYPE_INCOMPLETE
 );
 
 use constant MATCH_ALGORITHM_FULL      => 'FULL';
@@ -50,10 +48,23 @@ our @AS = qw(
    testSyn2Sig
    testSynASig
    dbFile
+   keepPcap
    _dump
+   _db
+   _tPatternBinary
+   _tPatternTcpFlags
+   _tPatternTcpWindow
+   _tPatternTcpOptions
+   _tPatternTcpMss
+   _tSystemClass
+   _tVendor
+   _tOs
+   _tOsVersion
+   _tOsVersionChildren
 );
 our @AA = qw(
    osfps
+   _signatures
 );
 __PACKAGE__->buildAccessorsScalar(\@AS);
 __PACKAGE__->buildAccessorsArray(\@AA);
@@ -83,52 +94,114 @@ Go to http://www.gomor.org/ to know more.
 
 =cut
 
-my $db;
-
-my $tSignature;
-my $tIpVersion;
-my $tPatternBinary;
-my $tPatternTcpFlags;
-my $tPatternTcpWindow;
-my $tPatternTcpOptions;
-my $tPatternTcpMss;
-my $tSystemClass;
-my $tVendor;
-my $tOs;
-my $tOsVersion;
-my $tOsVersionChildren;
-
 sub new {
    my $self = shift->SUPER::new(
-      offline => 0,
-      wait    => 3,
-      retry   => 3,
-      found   => 0,
-      passive => 0,
-      h2Match => 0,
-      osfps   => [],
+      offline  => 0,
+      wait     => 3,
+      retry    => 3,
+      found    => 0,
+      passive  => 0,
+      h2Match  => 0,
+      osfps    => [],
+      dbFile   => '/usr/local/share/sinfp/sinfp.db',
+      keepPcap => 0,
+      _db      => 0,
       @_,
    );
 
-   $db = DBIx::SQLite::Simple->new(db => $self->dbFile)
-      or die("Can't open db");
+   $self->_db(DBIx::SQLite::Simple->new(db => $self->dbFile))
+      or die("Can't open db: ". $self->dbFile. "\n");
 
-   $tSignature         = Net::SinFP::DB::Signature->new;
-   $tIpVersion         = Net::SinFP::DB::IpVersion->new;
-   $tPatternBinary     = Net::SinFP::DB::PatternBinary->new;
-   $tPatternTcpFlags   = Net::SinFP::DB::PatternTcpFlags->new;
-   $tPatternTcpWindow  = Net::SinFP::DB::PatternTcpWindow->new;
-   $tPatternTcpOptions = Net::SinFP::DB::PatternTcpOptions->new;
-   $tPatternTcpMss     = Net::SinFP::DB::PatternTcpMss->new;
-   $tSystemClass       = Net::SinFP::DB::SystemClass->new;
-   $tVendor            = Net::SinFP::DB::Vendor->new;
-   $tOs                = Net::SinFP::DB::Os->new;
-   $tOsVersion         = Net::SinFP::DB::OsVersion->new;
-   $tOsVersionChildren = Net::SinFP::DB::OsVersionChildren->new;
+   $self->_loadSignatures;
 
    $self;
 }
 
+sub _lookupOsInfos {
+   my $self = shift;
+   my $s = shift;
+
+   # Lookup values
+   $s->systemClass($self->_tSystemClass->getSystemClass($s->idSystemClass));
+   $s->vendor($self->_tVendor->getVendor($s->idVendor));
+   $s->os($self->_tOs->getOs($s->idOs));
+   $s->osVersion($self->_tOsVersion->getOsVersion($s->idOsVersion));
+
+   my $osVersionChildren = $self->_tOsVersionChildren->select(
+      idSignature => $s->idSignature,
+   );
+
+   if (@$osVersionChildren) {
+      my @osVersion;
+      push @osVersion, $self->_tOsVersion->getOsVersion($_->idOsVersion)
+         for @$osVersionChildren;
+      $s->osVersionChildren(\@osVersion);
+   }
+   else {
+      $s->osVersionChildren([]);
+   }
+}
+
+sub _lookupPatterns {
+   my $self = shift;
+   my $s = shift;
+
+   for my $t ('1', '2', '3') {
+      for my $m ('PatternBinary', 'PatternTcpFlags', 'PatternTcpWindow', 
+                 'PatternTcpOptions', 'PatternTcpMss') {
+         my $table = '_t'.$m;
+         my $g = $m;
+         $g =~ s/^Pattern/get/;
+         my $m1   = 't'.$t.$m;
+         my $m2   = $g;
+         my $m3   = 'idT'.$t.$m;
+         my $m1h1 = 't'.$t.$m.'H1';
+         my $m2h1 = $g.'H1';
+         my $m1h2 = 't'.$t.$m.'H2';
+         my $m2h2 = $g.'H2';
+         $s->$m1  ($self->$table->$m2  ($s->$m3));
+         $s->$m1h1($self->$table->$m2h1($s->$m3));
+         $s->$m1h2($self->$table->$m2h2($s->$m3));
+      }
+   }
+}
+
+sub _loadSignatures {
+   my $self = shift;
+
+   # Tables only used locally
+   my $tSignature = Net::SinFP::DB::Signature->new;
+   my $tIpVersion = Net::SinFP::DB::IpVersion->new;
+
+   # Tables used in other methods
+   $self->_tPatternBinary    (Net::SinFP::DB::PatternBinary->new);
+   $self->_tPatternTcpFlags  (Net::SinFP::DB::PatternTcpFlags->new);
+   $self->_tPatternTcpWindow (Net::SinFP::DB::PatternTcpWindow->new);
+   $self->_tPatternTcpOptions(Net::SinFP::DB::PatternTcpOptions->new);
+   $self->_tPatternTcpMss    (Net::SinFP::DB::PatternTcpMss->new);
+
+   $self->_tSystemClass      (Net::SinFP::DB::SystemClass->new);
+   $self->_tVendor           (Net::SinFP::DB::Vendor->new);
+   $self->_tOs               (Net::SinFP::DB::Os->new);
+   $self->_tOsVersion        (Net::SinFP::DB::OsVersion->new);
+   $self->_tOsVersionChildren(Net::SinFP::DB::OsVersionChildren->new);
+
+   my $idIpVersion;
+   my ($class) = ref($self) =~ /^(?:.*::)?(.*)/;
+
+   ($class eq 'SinFP4')
+      ? ($idIpVersion = $tIpVersion->getIdIpVersion('IPv4'))
+      : ($idIpVersion = $tIpVersion->getIdIpVersion('IPv6'));
+   my $signatures = $tSignature->select(idIpVersion => $idIpVersion);
+   die("Unable to load signatures from sinfp.db.\n".
+       "Try installing latest DBD::SQLite module.\n")
+      unless scalar @$signatures;
+
+   $self->_lookupPatterns($_) for @$signatures;
+
+   $self->_signatures($signatures);
+}
+
 sub _printPassive {
    my $self = shift;
    my ($frame) = @_;
@@ -206,8 +279,6 @@ sub startOnlinePassive {
          $self->_printPassive($frame);
       }
    }
-
-   $dump->stop;
 }
 
 sub startOfflinePassive {
@@ -256,7 +327,7 @@ sub startOnline {
 
    my $dump = Net::Packet::Dump->new(
       file            => $file,
-      unlinkOnDestroy => 0,
+      unlinkOnDestroy => $self->keepPcap ? 0 : 1,
       overwrite       => 1,
       timeoutOnNext   => $self->wait,
       callStart       => 0,
@@ -288,15 +359,13 @@ sub startOnline {
             $self->testSynAPkt->recv;
          }
 
-         do { $dump->stop; return } if $self->testSyn1Pkt->reply
-                                    && $self->testSyn2Pkt->reply
-                                    && $self->testSynAPkt->reply;
+         return if $self->testSyn1Pkt->reply
+                && $self->testSyn2Pkt->reply
+                && $self->testSynAPkt->reply;
       }
 
       $Env->dump->timeout(0);
    }
-
-   $dump->stop;
 }
 
 sub _startOfflineGetDump {
@@ -323,8 +392,6 @@ sub _startOfflineGetResponses {
    $self->testSyn1Pkt->recv if $self->testSyn1Pkt;
    $self->testSyn2Pkt->recv if $self->testSyn2Pkt;
    $self->testSynAPkt->recv if $self->testSynAPkt;
-
-   $self->_dump->stop;
 }
 
 sub _buildSigFromOptions {
@@ -400,18 +467,6 @@ sub _matchSig {
    my $self = shift;
    my ($type) = @_;
 
-   my ($class) = ref($self) =~ /^(?:.*::)?(.*)/;
-
-   my $signatures;
-   if ($class eq 'SinFP6') {
-      my $idIpVersion = $tIpVersion->getIdIpVersion('IPv6');
-      $signatures = $tSignature->select(idIpVersion => $idIpVersion);
-   }
-   else {
-      my $idIpVersion = $tIpVersion->getIdIpVersion('IPv4');
-      $signatures = $tSignature->select(idIpVersion => $idIpVersion);
-   }
-
    my $s1 = $self->testSyn1Sig if $self->testSyn1Pkt
                                && $self->testSyn1Pkt->reply;
    my $s2 = $self->testSyn2Sig if $self->testSyn2Pkt
@@ -419,133 +474,110 @@ sub _matchSig {
    my $sA = $self->testSynASig if $self->testSynAPkt
                                && $self->testSynAPkt->reply;
 
-   for my $s (@{$signatures}) {
+   for my $s ($self->_signatures) {
       my $t1;
       my $t2;
       my $t3;
       if ($type eq SIGNATURE_TYPE_EXACT) {
-         $t1 =
-            $tPatternBinary->getBinary($s->idT1PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlags($s->idT1PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindow($s->idT1PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptions($s->idT1PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMss($s->idT1PatternTcpMss)
-         ;
-         $t2 =
-            $tPatternBinary->getBinary($s->idT2PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlags($s->idT2PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindow($s->idT2PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptions($s->idT2PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMss($s->idT2PatternTcpMss)
-         ;
-         $t3 =
-            $tPatternBinary->getBinary($s->idT3PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlags($s->idT3PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindow($s->idT3PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptions($s->idT3PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMss($s->idT3PatternTcpMss)
-         ;
+         $t1 = $s->t1PatternBinary.' '.$s->t1PatternTcpFlags.' '.
+               $s->t1PatternTcpWindow.' '.$s->t1PatternTcpOptions.' '.
+               $s->t1PatternTcpMss;
+         $t2 = $s->t2PatternBinary.' '.$s->t2PatternTcpFlags.' '.
+               $s->t2PatternTcpWindow.' '.$s->t2PatternTcpOptions.' '.
+               $s->t2PatternTcpMss;
+         $t3 = $s->t3PatternBinary.' '.$s->t3PatternTcpFlags.' '.
+               $s->t3PatternTcpWindow.' '.$s->t3PatternTcpOptions.' '.
+               $s->t3PatternTcpMss;
          $s->signatureType(SIGNATURE_TYPE_EXACT);
       }
       elsif ($type eq SIGNATURE_TYPE_HEURISTIC1) {
-         $t1 =
-            $tPatternBinary->getBinaryH1($s->idT1PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlagsH1($s->idT1PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindowH1($s->idT1PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptionsH1($s->idT1PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMssH1($s->idT1PatternTcpMss)
-         ;
-         $t2 =
-            $tPatternBinary->getBinaryH1($s->idT2PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlagsH1($s->idT2PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindowH1($s->idT2PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptionsH1($s->idT2PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMssH1($s->idT2PatternTcpMss)
-         ;
-         $t3 =
-            $tPatternBinary->getBinaryH1($s->idT3PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlagsH1($s->idT3PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindowH1($s->idT3PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptionsH1($s->idT3PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMssH1($s->idT3PatternTcpMss)
-         ;
+         $t1 = $s->t1PatternBinaryH1.' '.$s->t1PatternTcpFlagsH1.' '.
+               $s->t1PatternTcpWindowH1.' '.$s->t1PatternTcpOptionsH1.' '.
+               $s->t1PatternTcpMssH1;
+         $t2 = $s->t2PatternBinaryH1.' '.$s->t2PatternTcpFlagsH1.' '.
+               $s->t2PatternTcpWindowH1.' '.$s->t2PatternTcpOptionsH1.' '.
+               $s->t2PatternTcpMssH1;
+         $t3 = $s->t3PatternBinaryH1.' '.$s->t3PatternTcpFlagsH1.' '.
+               $s->t3PatternTcpWindowH1.' '.$s->t3PatternTcpOptionsH1.' '.
+               $s->t3PatternTcpMssH1;
          $s->signatureType(SIGNATURE_TYPE_HEURISTIC1);
       }
       elsif ($type eq SIGNATURE_TYPE_HEURISTIC2) {
-         $t1 =
-            $tPatternBinary->getBinaryH2($s->idT1PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlagsH2($s->idT1PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindowH2($s->idT1PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptionsH2($s->idT1PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMssH2($s->idT1PatternTcpMss)
-         ;
-         $t2 =
-            $tPatternBinary->getBinaryH2($s->idT2PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlagsH2($s->idT2PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindowH2($s->idT2PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptionsH2($s->idT2PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMssH2($s->idT2PatternTcpMss)
-         ;
-         $t3 =
-            $tPatternBinary->getBinaryH2($s->idT3PatternBinary).' '.
-            $tPatternTcpFlags->getTcpFlagsH2($s->idT3PatternTcpFlags).' '.
-            $tPatternTcpWindow->getTcpWindowH2($s->idT3PatternTcpWindow).' '.
-            $tPatternTcpOptions->getTcpOptionsH2($s->idT3PatternTcpOptions).' '.
-            $tPatternTcpMss->getTcpMssH2($s->idT3PatternTcpMss)
-         ;
+         $t1 = $s->t1PatternBinaryH2.' '.$s->t1PatternTcpFlagsH2.' '.
+               $s->t1PatternTcpWindowH2.' '.$s->t1PatternTcpOptionsH2.' '.
+               $s->t1PatternTcpMssH2;
+         $t2 = $s->t2PatternBinaryH2.' '.$s->t2PatternTcpFlagsH2.' '.
+               $s->t2PatternTcpWindowH2.' '.$s->t2PatternTcpOptionsH2.' '.
+               $s->t2PatternTcpMssH2;
+         $t3 = $s->t3PatternBinaryH2.' '.$s->t3PatternTcpFlagsH2.' '.
+               $s->t3PatternTcpWindowH2.' '.$s->t3PatternTcpOptionsH2.' '.
+               $s->t3PatternTcpMssH2;
          $s->signatureType(SIGNATURE_TYPE_HEURISTIC2);
       }
 
+      # In passive mode, the SYN2 test is not our own, so timestamp is not 
+      # built as we want. We rewrite it to be able to match.
       if ($self->passive) {
          $t2 =~ s/44454144/......../;
       }
 
-      # Lookup values
-      $s->systemClass($tSystemClass->getSystemClass($s->idSystemClass));
-      $s->vendor($tVendor->getVendor($s->idVendor));
-      $s->os($tOs->getOs($s->idOs));
-      $s->osVersion($tOsVersion->getOsVersion($s->idOsVersion));
-
-      my $osVersionChildren = $tOsVersionChildren->select(
-         idSignature => $s->idSignature,
-      );
-
-      if (@$osVersionChildren) {
-         my @osVersion;
-         push @osVersion, $tOsVersion->getOsVersion($_->idOsVersion)
-            for @$osVersionChildren;
-         $s->osVersionChildren(\@osVersion);
-      }
-      else {
-         $s->osVersionChildren([]);
-      }
-
-      if ($s1 && $s2 && $sA
-      &&  $s1 =~ /^$t1$/
-      &&  $s2 =~ /^$t2$/
-      &&  $sA =~ /^$t3$/) {
+      # Matching is done here
+      if (($s1 && $s1 =~ /^$t1$/)
+      &&  ($s2 && $s2 =~ /^$t2$/)
+      &&  ($sA && $sA =~ /^$t3$/)) {
          $s->matchAlgorithm(MATCH_ALGORITHM_FULL);
+         $self->_lookupOsInfos($s);
          $self->_addResult($s);
-         $self->found($self->found + 1);
+         $self->found(1);
       }
-      elsif ($s1 && $s2 && ! $sA
-      &&     $s1 =~ /^$t1$/
-      &&     $s2 =~ /^$t2$/) {
-         $s->matchAlgorithm(MATCH_ALGORITHM_TWO);
+      elsif (($s1 && $s1 =~ /^$t1$/)
+         &&  ($s2 && $s2 =~ /^$t2$/)) {
+         $s->matchAlgorithm(MATCH_ALGORITHM_TWO); # Firewalled system
+         $self->_lookupOsInfos($s);
          $self->_addResult($s);
-         $self->found($self->found + 1);
+         $self->found(1);
       }
-      elsif (! $s1 && $s2 && ! $sA
-      &&     $s2 =~ /^$t2$/) {
+      elsif ($s2 && $s2 =~ /^$t2$/) { # Match only with test 2
          $s->matchAlgorithm(MATCH_ALGORITHM_ONE);
+         $self->_lookupOsInfos($s);
          $self->_addResult($s);
-         $self->found($self->found + 1);
+         $self->found(1);
       }
    }
 
    $self->found;
 }
 
+sub _cleanFound {
+   my $self = shift;
+
+   my $betterAlgo = MATCH_ALGORITHM_ONE;
+   for ($self->osfps) {
+      if ($_->matchAlgorithm eq MATCH_ALGORITHM_FULL) {
+         $betterAlgo = MATCH_ALGORITHM_FULL;
+         last;
+      }
+      elsif ($_->matchAlgorithm eq MATCH_ALGORITHM_TWO) {
+         $betterAlgo = MATCH_ALGORITHM_TWO;
+      }
+   }
+
+   my $h = {
+      MATCH_ALGORITHM_FULL() => 3,
+      MATCH_ALGORITHM_TWO()  => 2,
+      MATCH_ALGORITHM_ONE()  => 1,
+   };
+
+   my @keep = ();
+   for ($self->osfps) {
+      if ($h->{$_->matchAlgorithm} >= $h->{$betterAlgo}) {
+         push @keep, $_;
+      }
+   }
+   $self->osfps(\@keep);
+   $self->found(scalar @keep);
+}
+
 sub matchSigExact      { shift->_matchSig(SIGNATURE_TYPE_EXACT)      }
 sub matchSigHeuristic1 { shift->_matchSig(SIGNATURE_TYPE_HEURISTIC1) }
 sub matchSigHeuristic2 { shift->_matchSig(SIGNATURE_TYPE_HEURISTIC2) }
@@ -564,6 +596,11 @@ sub matchOsfps {
       #Remove testSynA (potentially firewall crafted), and retry
       $self->testSynAPkt && $self->testSynAPkt->reply(undef);
    }
+
+   # Keep only better MATCH_ALGORITHM from all results found
+   # Otherwise, it is possible to have a FULL match, and a FIREWALLED 
+   # match, and we do not want that.
+   $self->_cleanFound;
 }
 
 sub _getIpVersion {
@@ -632,9 +669,10 @@ sub printResultsOnlyOs {
 }
 
 sub DESTROY {
-   $Env->dump->stop if $Env->dump->isRunning;
-   $db->close       if $db;
-   exit(0);
+   my $self = shift;
+   $self->_dump->stop if $self->_dump && $self->_dump->isRunning;
+   $self->_db->close  if $self->_db;
+   return(0);
 }
 
 =head1 AUTHOR
@@ -643,7 +681,7 @@ Patrice E<lt>GomoRE<gt> Auffret
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (c) 2005, Patrice E<lt>GomoRE<gt> Auffret
+Copyright (c) 2005-2006, Patrice E<lt>GomoRE<gt> Auffret
 
 You may distribute this module under the terms of the Artistic license.
 See Copying file in the source distribution archive.
